package com.sun.org.apache.xalan.internal.xsltc.compiler;

import java.util.Vector;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.MethodType;
import com.sun.org.apache.bcel.internal.generic.InstructionHandle;
import com.sun.org.apache.bcel.internal.generic.IFEQ;
import com.sun.org.apache.bcel.internal.generic.BranchHandle;
import com.sun.org.apache.bcel.internal.generic.ConstantPoolGen;
import com.sun.org.apache.bcel.internal.generic.GOTO_W;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.NodeSetType;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.BooleanType;
import com.sun.org.apache.bcel.internal.generic.InstructionList;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.ErrorMsg;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.MethodGenerator;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.ClassGenerator;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.TypeCheckError;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.Type;

abstract class Expression extends SyntaxTreeNode
{
    protected Type _type;
    protected FlowList _trueList;
    protected FlowList _falseList;
    
    Expression() {
        this._trueList = new FlowList();
        this._falseList = new FlowList();
    }
    
    public Type getType() {
        return this._type;
    }
    
    @Override
    public abstract String toString();
    
    public boolean hasPositionCall() {
        return false;
    }
    
    public boolean hasLastCall() {
        return false;
    }
    
    public Object evaluateAtCompileTime() {
        return null;
    }
    
    @Override
    public Type typeCheck(SymbolTable stable) throws TypeCheckError {
        return this.typeCheckContents(stable);
    }
    
    @Override
    public void translate(ClassGenerator classGen, MethodGenerator methodGen) {
        ErrorMsg msg = new ErrorMsg("NOT_IMPLEMENTED_ERR", this.getClass(), this);
        this.getParser().reportError(2, msg);
    }
    
    public final InstructionList compile(ClassGenerator classGen, MethodGenerator methodGen) {
        InstructionList save = methodGen.getInstructionList();
        InstructionList result;
        methodGen.setInstructionList(result = new InstructionList());
        this.translate(classGen, methodGen);
        methodGen.setInstructionList(save);
        return result;
    }
    
    public void translateDesynthesized(ClassGenerator classGen, MethodGenerator methodGen) {
        this.translate(classGen, methodGen);
        if ((this._type instanceof BooleanType)) {
            this.desynthesize(classGen, methodGen);
        }
    }
    
    public void startIterator(ClassGenerator classGen, MethodGenerator methodGen) {
        if (!(this._type instanceof NodeSetType)) {
            return;
        }
        Expression expr = this;
        if ((expr instanceof CastExpr)) {
            expr = ((CastExpr)expr).getExpr();
        }
        if (!(expr instanceof VariableRefBase)) {
            InstructionList il = methodGen.getInstructionList();
            il.append(methodGen.loadContextNode());
            il.append(methodGen.setStartNode());
        }
        return;
    }
    
    public void synthesize(ClassGenerator classGen, MethodGenerator methodGen) {
        ConstantPoolGen cpg = classGen.getConstantPool();
        InstructionList il = methodGen.getInstructionList();
        this._trueList.backPatch(il.append(Expression.ICONST_1));
        BranchHandle truec = il.append(new GOTO_W(null));
        this._falseList.backPatch(il.append(Expression.ICONST_0));
        truec.setTarget(il.append(Expression.NOP));
    }
    
    public void desynthesize(ClassGenerator classGen, MethodGenerator methodGen) {
        InstructionList il = methodGen.getInstructionList();
        this._falseList.add(il.append(new IFEQ(null)));
    }
    
    public FlowList getFalseList() {
        return this._falseList;
    }
    
    public FlowList getTrueList() {
        return this._trueList;
    }
    
    public void backPatchFalseList(InstructionHandle ih) {
        this._falseList.backPatch(ih);
    }
    
    public void backPatchTrueList(InstructionHandle ih) {
        this._trueList.backPatch(ih);
    }
    
    public MethodType lookupPrimop(SymbolTable stable, String op, MethodType ctype) {
        MethodType result = null;
        Vector<Object> primop = stable.lookupPrimop(op);
        if (primop != null) {
            int n = primop.size();
            int minDistance = Integer.MAX_VALUE;
            for (int i = 0; i < n; ++i) {
                MethodType ptype = (MethodType)primop.elementAt(i);
                if (ptype.argsCount() == ctype.argsCount()) {
                    if (result == null) {
                        result = ptype;
                    }
                    int distance = ctype.distanceTo(ptype);
                    if (distance < minDistance) {
                        minDistance = distance;
                        result = ptype;
                    }
                }
            }
            return result;
        }
        return result;
    }
}
